#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
import threading
from typing import List

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
from fabric.actor.core.apis.i_database import IDatabase
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_reservation import IReservation, ReservationCategory
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import DatabaseException
from fabric.actor.core.kernel.slice import SliceTypes
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.db.psql_database import PsqlDatabase


class ActorDatabase(IDatabase):
    def __init__(self, *, user: str, password: str, database: str, db_host: str, logger):
        self.user = user
        self.password = password
        self.database = database
        self.db_host = db_host
        self.db = PsqlDatabase(user=self.user, password=self.password, database=self.database, db_host=self.db_host,
                               logger=logger)
        self.actor_name = None
        self.actor_id = None
        self.initialized = False
        self.logger = None
        self.reset_state = False
        self.lock = threading.Lock()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['db']
        del state['actor_name']
        del state['actor_id']
        del state['initialized']
        del state['logger']
        del state['reset_state']
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.db = PsqlDatabase(user=self.user, password=self.password, database=self.database, db_host=self.db_host,
                               logger=None)
        self.actor_id = None
        self.actor_name = None
        self.initialized = False
        self.logger = None
        self.reset_state = False
        self.lock = threading.Lock()

    def set_logger(self, *, logger):
        self.logger = logger

    def set_reset_state(self, *, state: bool):
        self.reset_state = state

    def initialize(self):
        if not self.initialized:
            if self.actor_name is None:
                raise DatabaseException(Constants.not_specified_prefix.format("actor name"))
            self.initialized = True

    def actor_added(self):
        self.actor_id = self.get_actor_id_from_name(actor_name=self.actor_name)
        if self.actor_id is None:
            raise DatabaseException(Constants.object_not_found.format("actor", self.actor_name))

    def revisit(self, *, actor: IActor, properties: dict):
        return

    def get_actor_id_from_name(self, *, actor_name: str) -> int:
        try:
            self.lock.acquire()
            actor = self.db.get_actor(name=actor_name)
            self.actor_id = actor['act_id']
            return self.actor_id
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_slice_id_from_guid(self, *, slice_id: ID) -> int:
        try:
            self.lock.acquire()
            slice_obj = self.db.get_slice(act_id=self.actor_id, slice_guid=str(slice_id))
            return slice_obj['slc_id']
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_slice_by_id(self, *, id: int) -> dict:
        try:
            self.lock.acquire()
            slice_obj = self.db.get_slice_by_id(act_id=self.actor_id, id=id)
            return slice_obj
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def add_slice(self, *, slice_object: ISlice):
        try:
            if self.get_slice(slice_id=slice_object.get_slice_id()) is not None:
                raise DatabaseException("Slice # {} already exists".format(slice_object.get_slice_id()))
            self.lock.acquire()
            properties = pickle.dumps(slice_object)
            self.db.add_slice(act_id=self.actor_id, slc_guid=str(slice_object.get_slice_id()),
                              slc_name=slice_object.get_name(),
                              slc_type=slice_object.get_slice_type().value,
                              slc_resource_type=str(slice_object.get_resource_type()),
                              properties=properties,
                              slc_graph_id=str(slice_object.get_graph_id()))
        finally:
            self.lock.release()

    def update_slice(self, *, slice_object: ISlice):
        try:
            self.lock.acquire()
            properties = pickle.dumps(slice_object)
            self.db.update_slice(act_id=self.actor_id,
                                 slc_guid=str(slice_object.get_slice_id()),
                                 slc_name=slice_object.get_name(),
                                 slc_type=slice_object.get_slice_type().value,
                                 slc_resource_type=str(slice_object.get_resource_type()),
                                 properties=properties,
                                 slc_graph_id=str(slice_object.get_graph_id()))
        finally:
            self.lock.release()

    def remove_slice(self, *, slice_id: ID):
        try:
            self.lock.acquire()
            self.db.remove_slice(slc_guid=str(slice_id))
        finally:
            self.lock.release()

    def get_slice(self, *, slice_id: ID) -> dict:
        try:
            return self.db.get_slice(act_id=self.actor_id, slice_guid=str(slice_id))
        except Exception as e:
            self.logger.error(e)
        return None

    def get_slices(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_slices(act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_inventory_slices(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_slices_by_type(act_id=self.actor_id, slc_type=SliceTypes.InventorySlice.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_client_slices(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_slices_by_types(act_id=self.actor_id,
                                               slc_type1=SliceTypes.ClientSlice.value,
                                               slc_type2=SliceTypes.BrokerClientSlice.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_slice_by_resource_type(self, *, rtype: ResourceType) -> dict:
        try:
            self.lock.acquire()
            result_list = self.db.get_slices_by_resource_type(act_id=self.actor_id, slc_resource_type=str(rtype))
            if result_list is not None and len(result_list) > 0:
                return next(iter(result_list))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def add_reservation(self, *, reservation: IReservation):
        try:
            self.lock.acquire()
            self.logger.debug("Adding reservation {} to slice {}".format(reservation.get_reservation_id(),
                                                                         reservation.get_slice()))
            properties = pickle.dumps(reservation)
            self.db.add_reservation(act_id=self.actor_id,
                                    slc_guid=str(reservation.get_slice_id()),
                                    rsv_resid=str(reservation.get_reservation_id()),
                                    rsv_category=reservation.get_category().value,
                                    rsv_state=reservation.get_state().value,
                                    rsv_pending=reservation.get_pending_state().value,
                                    rsv_joining=reservation.get_join_state().value,
                                    properties=properties)
            self.logger.debug(
                "Reservation {} added to slice {}".format(reservation.get_reservation_id(), reservation.get_slice()))
        finally:
            self.lock.release()

    def update_reservation(self, *, reservation: IReservation):
        # Update the reservation only when there are changes to be reflected in database
        if not reservation.is_dirty():
            return
        reservation.clear_dirty()
        try:
            self.lock.acquire()
            self.logger.debug("Updating reservation {} in slice {}".format(reservation.get_reservation_id(),
                                                                           reservation.get_slice()))
            properties = pickle.dumps(reservation)
            self.db.update_reservation(act_id=self.actor_id,
                                       slc_guid=str(reservation.get_slice_id()),
                                       rsv_resid=str(reservation.get_reservation_id()),
                                       rsv_category=reservation.get_category().value,
                                       rsv_state=reservation.get_state().value,
                                       rsv_pending=reservation.get_pending_state().value,
                                       rsv_joining=reservation.get_join_state().value,
                                       properties=properties)

        finally:
            self.lock.release()

    def remove_reservation(self, *, rid: ID):
        try:
            self.lock.acquire()
            self.logger.debug("Removing reservation {}".format(rid))
            self.db.remove_reservation(rsv_resid=str(rid))
        finally:
            self.lock.release()

    def get_reservation(self, *, rid: ID) -> dict:
        try:
            self.lock.acquire()
            return self.db.get_reservation(act_id=self.actor_id, rsv_resid=str(rid))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_reservations_by_slice_id(self, *, slice_id: ID) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_slice_id(act_id=self.actor_id, slc_guid=str(slice_id))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_reservations_by_slice_id_state(self, *, slice_id: ID, state: int) -> list:
        try:
            return self.db.get_reservations_by_slice_id_state(act_id=self.actor_id, slc_guid=str(slice_id),
                                                              rsv_state=state)
        except Exception as e:
            self.logger.error(e)
        return None

    def get_client_reservations(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_2_category(act_id=self.actor_id,
                                                          rsv_cat1=ReservationCategory.Broker.value,
                                                          rsv_cat2=ReservationCategory.Authority.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_client_reservations_by_slice_id(self, *, slice_id: ID) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_slice_id_by_2_category(act_id=self.actor_id,
                                                                      slc_guid=str(slice_id),
                                                                      rsv_cat1=ReservationCategory.Broker.value,
                                                                      rsv_cat2=ReservationCategory.Authority.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_holdings(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_category(act_id=self.actor_id, rsv_cat=ReservationCategory.Client.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_holdings_by_slice_id(self, *, slice_id: ID) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_slice_id_by_category(act_id=self.actor_id, slc_guid=str(slice_id),
                                                                    rsv_cat=ReservationCategory.Client.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_broker_reservations(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_category(act_id=self.actor_id, rsv_cat=ReservationCategory.Broker.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_authority_reservations(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_category(act_id=self.actor_id,
                                                        rsv_cat=ReservationCategory.Authority.value)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_reservations(self) -> list:
        try:
            self.lock.acquire()
            self.logger.debug("Actor ID: {}".format(self.actor_id))
            return self.db.get_reservations(act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_reservations_by_state(self, *, state: int) -> list:
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_state(act_id=self.actor_id, rsv_state=state)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_reservations_by_rids(self, *, rid: List[str]):
        try:
            self.lock.acquire()
            return self.db.get_reservations_by_rids(act_id=self.actor_id, rsv_resid_list=rid)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def add_broker(self, *, broker: IBrokerProxy):
        try:
            self.lock.acquire()
            self.logger.debug("Adding broker {}({})".format(broker.get_name(), broker.get_guid()))
            properties = pickle.dumps(broker)
            self.db.add_proxy(act_id=self.actor_id, prx_name=broker.get_name(), properties=properties)
        finally:
            self.lock.release()

    def update_broker(self, *, broker: IBrokerProxy):
        try:
            self.lock.acquire()
            self.logger.debug("Updating broker {}({})".format(broker.get_name(), broker.get_guid()))
            properties = pickle.dumps(broker)
            self.db.update_proxy(act_id=self.actor_id, prx_name=broker.get_name(), properties=properties)
        finally:
            self.lock.release()

    def remove_broker(self, *, broker: IBrokerProxy):
        try:
            self.lock.acquire()
            self.logger.debug("Removing broker {}({})".format(broker.get_name(), broker.get_guid()))
            self.db.remove_proxy(act_id=self.actor_id, prx_name=broker.get_name())
        finally:
            self.lock.release()

    def set_actor_name(self, *, name: str):
        self.actor_name = name

    def get_brokers(self) -> list:
        try:
            self.lock.acquire()
            return self.db.get_proxies(act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def add_delegation(self, *, delegation: IDelegation):
        try:
            self.lock.acquire()
            self.logger.debug("Adding delegation {} to slice {}".format(delegation.get_delegation_id(),
                                                                        delegation.get_slice_id()))

            slice_object = self.get_slice(slice_id=delegation.get_slice_id())
            if slice_object is None:
                raise DatabaseException("Slice with id: {} not found".format(delegation.get_slice_id()))

            properties = pickle.dumps(delegation)
            self.db.add_delegation(dlg_act_id=self.actor_id,
                                   dlg_slc_id=slice_object['slc_id'],
                                   dlg_graph_id=str(delegation.get_delegation_id()),
                                   dlg_state=delegation.get_state().value,
                                   properties=properties)
            self.logger.debug(
                "Delegation {} added to slice {}".format(delegation.get_delegation_id(),
                                                         delegation.get_slice_id()))
        finally:
            self.lock.release()

    def update_delegation(self, *, delegation: IDelegation):
        # Update the delegation only when there are changes to be reflected in database
        if not delegation.is_dirty():
            return
        delegation.clear_dirty()
        try:
            self.lock.acquire()
            self.logger.debug("Updating delegation {} in slice {}".format(delegation.get_delegation_id(),
                                                                          delegation.get_slice_id()))
            properties = pickle.dumps(delegation)
            self.db.update_delegation(dlg_act_id=self.actor_id,
                                      dlg_graph_id=str(delegation.get_delegation_id()),
                                      dlg_state=delegation.get_state().value,
                                      properties=properties)
        finally:
            self.lock.release()

    def remove_delegation(self, *, dlg_graph_id: ID):
        try:
            self.lock.acquire()
            self.logger.debug("Removing delegation {}".format(dlg_graph_id))
            self.db.remove_delegation(dlg_graph_id=str(dlg_graph_id))
        finally:
            self.lock.release()

    def get_delegation(self, *, dlg_graph_id: ID) -> dict:
        try:
            self.lock.acquire()
            return self.db.get_delegation(dlg_act_id=self.actor_id, dlg_graph_id=str(dlg_graph_id))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_delegations(self) -> list:
        try:
            self.lock.acquire()
            self.logger.debug("Actor ID: {}".format(self.actor_id))
            return self.db.get_delegations(dlg_act_id=self.actor_id)
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None

    def get_delegations_by_slice_id(self, *, slice_id: ID) -> list:
        try:
            self.lock.acquire()
            return self.db.get_delegations_by_slice_id(dlg_act_id=self.actor_id, slc_guid=str(slice_id))
        except Exception as e:
            self.logger.error(e)
        finally:
            self.lock.release()
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.resource_vector import ResourceVector
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID


class ResourceBin:
    """
    ResourceBin represents a grouping of resources with identical properties.
    Each resource bin consists of one or more physical units with a given resource vector.
    Each resource bin has a unique identifier and a unique parent resource bin.
    """
    def __init__(self, *, parent_guid: ID = None, physical_units: int = None, vector: ResourceVector = None,
                 term: Term = None):
        self.guid = ID()
        self.parent_guid = parent_guid
        self.physical_units = physical_units
        self.vector = vector
        self.term = term

    def get_guid(self) -> ID:
        """
        Get Guid
        @return guid
        """
        return self.guid

    def get_parent_guid(self) -> ID:
        """
        Get Parent Guid
        @return parent guid
        """
        return self.parent_guid

    def get_physical_units(self) -> int:
        """
        Get Physical Unit count
        @return physical unit count
        """
        return self.physical_units

    def get_vector(self) -> ResourceVector:
        """
        Get resource vector
        @return resource vector
        """
        return self.vector

    def get_term(self) -> Term:
        """
        Return Term
        @return term
        """

        return self.term



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.apis.i_authority_policy import IAuthorityPolicy
from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.core.policy import Policy
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.delegation.resource_delegation import ResourceDelegation
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData


class AuthorityPolicy(Policy, IAuthorityPolicy):
    PropertySourceTicket = "sourceTicket"

    """
    Base class for all authority policy implementations.
    """
    def __init__(self, *, actor: IAuthority = None):
        super().__init__(actor=actor)

        self.tickets = None
        self.initialized = False
        self.delegations = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['actor']
        del state['clock']
        del state['initialized']
        del state['tickets']
        del state['delegations']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.actor = None
        self.clock = None
        self.initialized = False
        self.tickets = None
        self.delegations = None

    def initialize(self):
        if not self.initialized:
            super().initialize()
            self.tickets = {}
            self.initialized = True
            self.delegations = {}

    def revisit(self, *, reservation: IReservation):
        if isinstance(reservation, IClientReservation):
            self.donate_reservation(reservation=reservation)

    def revisit_delegation(self, *, delegation: IDelegation):
        self.donate_delegation(delegation=delegation)

    def donate_delegation(self, *, delegation: IDelegation):
        self.delegations[delegation.get_delegation_id()] = delegation
        # TODO need to determine how to populate controls and tickets

    def donate(self, *, resources: ResourceSet):
        return

    def donate_reservation(self, *, reservation: IClientReservation):
        rset = reservation.get_resources()
        rset.set_reservation_id(rid=reservation.get_reservation_id())

        self.logger.info("AuthorityPolicy donateTicket {}".format(rset))

        ticket_set = self.tickets.get(rset.get_type(), None)

        if ticket_set is None:
            ticket_set = set()

        ticket_set.add(rset)

        self.logger.debug("Adding tickets {} for type: {}".format(len(ticket_set), rset.get_type()))
        self.tickets[rset.get_type()] = ticket_set

    def eject(self, *, resources: ResourceSet):
        return

    def failed(self, *, resources: ResourceSet):
        return

    def unavailable(self, *, resources: ResourceSet) -> int:
        return 0

    def available(self, *, resources: ResourceSet):
        return

    def freed(self, *, resources: ResourceSet):
        return

    def recovered(self, *, resources: ResourceSet):
        return

    def release(self, *, resources: ResourceSet):
        return

    def bind_delegation(self, *, delegation: IDelegation) -> bool:
        result = False

        if delegation.get_delegation_id() not in self.delegations:
            self.delegations[delegation.get_delegation_id()] = delegation
            result = True

        return result

    def bind(self, *, reservation: IBrokerReservation) -> bool:
        requested = reservation.get_requested_resources()
        ticket_set = self.tickets.get(requested.get_type(), None)
        if ticket_set is None or len(ticket_set) == 0:
            self.error(message="bindTicket: no tickets available for the requested resource type {}".format(
                requested.get_type()))

        #  We need a calendar to tell us what we have exported from each source
        #  ticket. For now, just take the first element from the set and export
        #  from it
        #  Also: it should be possible to specify a source ticket to use for a
        #  given export. This can be a property on the AgentReservation (say:
        #  slice and reservation id to use). Have to make sure that the resource set has
        #  the slice and the reservation id.
        rid = None
        if reservation.get_requested_resources().get_request_properties() is not None:
            temp = reservation.get_requested_resources().get_request_properties().get(self.PropertySourceTicket, None)

            if temp is not None:
                rid = ID(uid=temp)

        ticket_found = None
        if rid is not None:
            for ticket in ticket_set:
                if ticket.get_reservation_id() is not None and ticket.get_reservation_id() == rid:
                    ticket_found = ticket

        else:
            if ticket_set is not None:
                ticket_found = ticket_set.__iter__().__next__()

        if ticket_found is None:
            self.error(message="bindticket: no tickets available")

        # If this is an elastic export, whittle it down to size if necessary.
        # Signal obvious export errors. This code is fragile/temporary: it
        # assumes that the request is for "now", and that there is just one
        # export per resource type. Other errors involving multiple exports per
        # type will be caught in the extract below, but we won't have a chance
        # to adjust if the request is elastic.
        units = ticket_found.get_resources().get_units()
        needed = requested.get_units()

        if needed > units:
            self.logger.error("insufficient units available to export to agent")
            needed = units

        delegation = self.actor.get_plugin().get_ticket_factory().make_delegation(
            units=needed, term=reservation.get_requested_term(), rtype=requested.get_type(),
            holder=reservation.get_client_auth_token().get_guid())

        mine = self.extract(source=ticket_found, delegation=delegation)
        reservation.set_approved(term=reservation.get_requested_term(), approved_resources=mine)
        return True

    def allocate(self, *, cycle: int):
        return

    def assign(self, *, cycle: int):
        return

    def correct_deficit(self, *, reservation: IAuthorityReservation):
        if reservation.get_resources() is None:
            return

        self.finish_correct_deficit(reservation=reservation)

    def finish_correct_deficit(self, *, reservation: IAuthorityReservation, rset: ResourceSet = None):
        """
        Finishes correcting a deficit.
        @param rset correction
        @param reservation reservation
        @raises Exception in case of error
        """
        # We could have a partial set if there's a shortage. Go ahead and
        # install it: we'll come back later for the rest if we return a null
        # term. Alternatively, we could release them and throw an error.
        if rset is None:
            self.log_warn(message="we either do not have resources to satisfy the request or "
                                  "the reservation has/will have a pending operation")
            return

        if rset.is_empty():
            reservation.set_pending_recover(pending_recover=False)
        else:
            reservation.get_resources().update(reservation=reservation, resource_set=rset)

    def extend_authority(self, *, reservation: IAuthorityReservation) -> bool:
        return False

    def extend_broker(self, *, reservation: IBrokerReservation) -> bool:
        return False

    def extend(self, *, reservation: IReservation, resources: ResourceSet, term: Term):
        return False

    def extract(self, *, source: ResourceSet, delegation: ResourceDelegation):
        """
        Creates a new resource set using the source and the specified delegation.
        @param source source
        @param delegation delegation
        @return created resource set
        @raises Exception in case of error
        """
        if source is None or delegation is None:
            self.error(message="Invalid Argument")

        self.logger.debug("extracting delegation: {}".format(delegation))

        rd = ResourceData()
        rd.resource_properties = source.get_resource_properties()
        extracted = ResourceSet(units=delegation.get_units(), rtype=delegation.get_resource_type(), rdata=rd)
        source_ticket = source.get_resources().get_ticket()

        new_ticket = self.actor.get_plugin().get_ticket_factory().make_ticket(delegation=delegation,
                                                                              source=source_ticket)

        cset = Ticket(ticket=new_ticket, plugin=source.get_resources().plugin,
                      authority=source.get_resources().authority)
        extracted.set_resources(cset=cset)
        return extracted



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, List

from fabric.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric.message_bus.messages.result_strings_avro import ResultStringsAvro
from fabric.message_bus.messages.result_avro import ResultAvro

if TYPE_CHECKING:
    from fabric.message_bus.messages.ticket_reservation_avro import TicketReservationAvro
    from fabric.actor.security.auth_token import AuthToken
    from fabric.message_bus.messages.result_string_avro import ResultStringAvro
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.message_bus.messages.reservation_mng import ReservationMng
    from fabric.actor.core.util.id import ID
    from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
    from fabric.message_bus.messages.proxy_avro import ProxyAvro
    from fabric.message_bus.messages.result_proxy_avro import ResultProxyAvro
    from fabric.message_bus.messages.result_pool_info_avro import ResultPoolInfoAvro


class IClientActorManagementObject:
    @abstractmethod
    def add_reservation(self, *, reservation: TicketReservationAvro, caller: AuthToken) -> ResultStringAvro:
        """
        Add a reservation
        @param reservation : reservation to be added
        @param caller: caller
        @return reservation id on success and error code on failure (ResultStringAvro)
        """

    @abstractmethod
    def add_reservations(self, *, reservations: List[TicketReservationAvro], caller: AuthToken) -> ResultStringsAvro:
        """
        Add reservations
        @param reservations : reservations to be added
        @param caller: caller
        @return reservation ids on success and error code on failure (ResultStringsAvro)
        """

    @abstractmethod
    def extend_reservation(self, *, reservation: id, new_end_time: datetime, new_units: int,
                           new_resource_type: ResourceType, request_properties: dict,
                           config_properties: dict, caller: AuthToken) -> ResultAvro:
        """
        Extend a reservation
        @param reservation : reservation to be extended
        @param new_end_time: new end time
        @param new_resource_type: new resource type
        @param config_properties: config_properties
        @param caller: caller
        @return success or failure status
        """

    @abstractmethod
    def demand_reservation(self, *, reservation: ReservationMng, caller: AuthToken) -> ResultStringAvro:
        """
        Demand a reservation
        @param reservation : reservation to be demanded
        @param caller: caller
        @return reservation id on success and error code on failure (ResultStringAvro)
        """

    @abstractmethod
    def demand_reservation_rid(self, *, rid: ID, caller: AuthToken) -> ResultAvro:
        """
        Demand a reservation by reservation id
        @param rid : reservation id of reservation to be demanded
        @param caller: caller
        @return success or failure status
        """

    @abstractmethod
    def get_brokers(self, *, caller: AuthToken, broker_id: ID = None, id_token: str = None) -> ResultProxyAvro:
        """
        Get all brokers
        @param caller: caller
        @param broker_id : broker_id
        @param id_token: id token
        @return list of all the brokers
        """

    @abstractmethod
    def add_broker(self, *, broker: ProxyAvro, caller: AuthToken) -> ResultAvro:
        """
        Add a broker
        @param broker: broker_proxy to be added
        @param caller: caller
        @return success or failure status
        """

    @abstractmethod
    def get_pool_info(self, *, broker: ID, caller: AuthToken, id_token: str) -> ResultPoolInfoAvro:
        """
        Get Pool Info
        @param broker : broker ID
        @param caller: caller
        @param id_token: str
        @return pool information
        """

    @abstractmethod
    def claim_delegations(self, *, broker: ID, did: ID, caller: AuthToken,
                          id_token: str = None) -> ResultDelegationAvro:
        """
        Claim delegations
        @param broker : broker ID
        @param did : delegations ID
        @param caller: caller
        @param id_token: id token
        @return ResultDelegationAvro
        """

    @abstractmethod
    def reclaim_delegations(self, *, broker: ID, did: ID, caller: AuthToken,
                            id_token: str = None) -> ResultDelegationAvro:
        """
        ReClaim delegations
        @param broker : broker ID
        @param did : delegations ID
        @param caller: caller
        @param id_token: id token
        @return ResultDelegationAvro
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_concrete_set import IConcreteSet
from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.delegation.broker_delegation_factory import BrokerDelegationFactory
from fabric.actor.core.kernel.client_reservation_factory import ClientReservationFactory
from fabric.actor.core.kernel.incoming_delegation_rpc import IncomingDelegationRPC
from fabric.actor.core.kernel.incoming_failed_rpc import IncomingFailedRPC
from fabric.actor.core.kernel.incoming_query_rpc import IncomingQueryRPC
from fabric.actor.core.kernel.incoming_rpc import IncomingRPC
from fabric.actor.core.kernel.incoming_reservation_rpc import IncomingReservationRPC
from fabric.actor.core.kernel.rpc_manager_singleton import RPCManagerSingleton
from fabric.actor.core.kernel.rpc_request_type import RPCRequestType
from fabric.actor.core.proxies.kafka.kafka_retun import KafkaReturn
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.delegation_avro import DelegationAvro
from fabric.message_bus.messages.reservation_avro import ReservationAvro
from fabric.message_bus.messages.message import IMessageAvro
from fabric.message_bus.messages.update_delegation_avro import UpdateDelegationAvro

if TYPE_CHECKING:
    from fabric.actor.security.auth_token import AuthToken
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.message_bus.messages.failed_rpc_avro import FailedRpcAvro
    from fabric.message_bus.messages.query_avro import QueryAvro
    from fabric.message_bus.messages.query_result_avro import QueryResultAvro
    from fabric.message_bus.messages.update_lease_avro import UpdateLeaseAvro
    from fabric.message_bus.messages.update_ticket_avro import UpdateTicketAvro
    from fabric.actor.core.apis.i_actor import IActor


class ActorService:
    def __init__(self, *, actor: IActor):
        self.actor = actor
        self.logger = self.actor.get_logger()

    def get_callback(self, *, kafka_topic: str, auth: AuthToken):
        return KafkaReturn(kafka_topic=kafka_topic, identity=auth, logger=self.actor.get_logger())

    def get_concrete(self, *, reservation: ReservationAvro) -> IConcreteSet:
        encoded = reservation.resource_set.concrete
        try:
            decoded = pickle.loads(encoded)
            return decoded
        except Exception as e:
            self.logger.error("Exception occurred while decoding {}".format(e))
        return None

    def pass_client(self, *, reservation: ReservationAvro) -> IClientReservation:
        slice_obj = Translate.translate_slice(slice_id=reservation.slice.guid, slice_name=reservation.slice.slice_name)
        term = Translate.translate_term_from_avro(term=reservation.term)

        resource_set = Translate.translate_resource_set_from_avro(rset=reservation.resource_set)
        resource_set.set_resources(cset=self.get_concrete(reservation=reservation))

        return ClientReservationFactory.create(rid=ID(uid=reservation.reservation_id), resources=resource_set, term=term,
                                               slice_object=slice_obj, actor=self.actor)

    def pass_client_delegation(self, *, delegation: DelegationAvro) -> IDelegation:
        slice_obj = Translate.translate_slice(slice_id=delegation.slice.guid, slice_name=delegation.slice.slice_name)

        dlg = BrokerDelegationFactory.create(did=delegation.get_delegation_id(), slice_id=slice_obj.get_slice_id(),
                                             broker=None)
        dlg.restore(actor=self.actor, slice_obj=slice_obj, logger=self.logger)
        dlg.load_graph(graph_str=delegation.graph)
        return dlg

    def do_dispatch(self, *, rpc: IncomingRPC):
        try:
            RPCManagerSingleton.get().dispatch_incoming(actor=self.actor, rpc=rpc)
        except Exception as e:
            msg = "An error occurred while dispatching request"
            self.logger.error("{} e={}".format(msg, e))
            raise e

    def query(self, *, request: QueryAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            query = request.properties
            callback = self.get_callback(kafka_topic=request.callback_topic, auth=auth_token)
            rpc = IncomingQueryRPC(request_type=RPCRequestType.Query, message_id=ID(uid=request.get_message_id()),
                                   query=query, caller=auth_token, callback=callback, id_token=request.get_id_token())
        except Exception as e:
            self.logger.error("Invalid query request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def query_result(self, *, request: QueryResultAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            query = request.properties
            rpc = IncomingQueryRPC(request_type=RPCRequestType.QueryResult, message_id=ID(uid=request.get_message_id()),
                                   query=query, caller=auth_token, request_id=ID(uid=request.request_id), id_token=None)
        except Exception as e:
            self.logger.error("Invalid query_result request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def update_lease(self, *, request: UpdateLeaseAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_client(reservation=request.reservation)
            udd = Translate.translate_udd_from_avro(udd=request.update_data)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id), request_type=RPCRequestType.UpdateLease,
                                         reservation=rsvn, update_data=udd, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid update_lease request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def update_ticket(self, *, request: UpdateTicketAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            rsvn = self.pass_client(reservation=request.reservation)
            udd = Translate.translate_udd_from_avro(udd=request.update_data)
            rpc = IncomingReservationRPC(message_id=ID(uid=request.message_id),
                                         request_type=RPCRequestType.UpdateTicket, reservation=rsvn, update_data=udd,
                                         caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid update_ticket request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def update_delegation(self, *, request: UpdateDelegationAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            dlg = self.pass_client_delegation(delegation=request.delegation)
            udd = Translate.translate_udd_from_avro(udd=request.update_data)
            rpc = IncomingDelegationRPC(message_id=ID(uid=request.message_id),
                                        request_type=RPCRequestType.UpdateDelegation, delegation=dlg, update_data=udd,
                                        caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid update_delegation request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def failed_rpc(self, *, request: FailedRpcAvro):
        rpc = None
        auth_token = Translate.translate_auth_from_avro(auth_avro=request.auth)
        try:
            failed_request_type = RPCRequestType(request.request_type)
            if request.reservation_id is not None and request.reservation_id != "":
                rpc = IncomingFailedRPC(message_id=ID(uid=request.message_id), failed_request_type=failed_request_type,
                                        request_id=request.request_id,
                                        failed_reservation_id=ID(uid=request.reservation_id),
                                        error_details=request.error_details, caller=auth_token)
            else:
                rpc = IncomingFailedRPC(message_id=ID(uid=request.message_id), failed_request_type=failed_request_type,
                                        request_id=request.request_id, failed_reservation_id=None,
                                        error_details=request.error_details, caller=auth_token)
        except Exception as e:
            self.logger.error("Invalid failedRequest request: {}".format(e))
            raise e
        self.do_dispatch(rpc=rpc)

    def process(self, *, message: IMessageAvro):
        if message.get_message_name() == IMessageAvro.query:
            self.query(request=message)
        elif message.get_message_name() == IMessageAvro.query_result:
            self.query_result(request=message)
        elif message.get_message_name() == IMessageAvro.update_lease:
            self.update_lease(request=message)
        elif message.get_message_name() == IMessageAvro.update_ticket:
            self.update_ticket(request=message)
        elif message.get_message_name() == IMessageAvro.update_delegation:
            self.update_delegation(request=message)
        else:
            self.logger.error("Unsupported message {}".format(message))
            raise ProxyException("Unsupported message {}".format(message.get_message_name()))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import pickle
from datetime import datetime
from typing import TYPE_CHECKING

from fabric.actor.core.common.exceptions import UnitException
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.notice import Notice

from fabric.actor.core.apis.i_concrete_set import IConcreteSet
from fabric.actor.core.core.unit import UnitState, Unit

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_authority_proxy import IAuthorityProxy
    from fabric.actor.core.apis.i_base_plugin import IBasePlugin
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.resource_type import ResourceType


class UnitSet(IConcreteSet):
    """
    Represents the unit in a reservation
    """
    def __init__(self, *, plugin: IBasePlugin, units: dict = None):
        self.units = units
        if self.units is None:
            self.units = {}
        self.reservation = None
        self.plugin = plugin
        self.logger = None
        if self.plugin is not None:
            self.logger = self.plugin.logger
        self.is_closed = False
        self.is_fresh = False
        self.released = None

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        del state['plugin']
        del state['reservation']
        del state['released']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None
        self.plugin = None
        self.reservation = None
        self.released = None

    def __str__(self):
        result = ""
        for unit in self.units:
            result += "[{}]".format(unit)
        return result

    def restore(self, *, plugin: IBasePlugin, reservation: IReservation):
        """
        Restore post read from database
        @param plugin plugin
        @param reservation reservation
        """
        self.plugin = plugin
        self.logger = plugin.get_logger()
        self.reservation = reservation

    def ensure_type(self, *, cset: IConcreteSet):
        """
        Validate the type of incoming concrete set
        @param cset cset
        """
        if not isinstance(cset, UnitSet):
            raise UnitException("Must be UnitSet")

    def add_unit(self, *, u: Unit):
        """
        Add a unit
        @parm u unit
        """
        if u.get_id() not in self.units:
            self.units[u.get_id()] = u

    def add(self, *, concrete_set: IConcreteSet, configure: bool):
        self.ensure_type(cset=concrete_set)

        self.add_from_dict(units=concrete_set.units)

        if configure:
            self.transfer_in_units(units=concrete_set.units)

    def add_from_dict(self, *, units: dict):
        """
        Add units from a dictionary
        @param units units
        """
        self.is_fresh = False
        for u in units.values():
            self.add_unit(u=u)

    def missing(self, *, units: dict) -> dict:
        """
        Find units not present in incoming units
        @param units incoming units
        """
        result = {}
        for u in self.units.values():
            if u.get_id() not in units:
                result[u.get_id()] = u
        return result

    def change(self, *, concrete_set: IConcreteSet, configure: bool):
        if not isinstance(concrete_set, UnitSet):
            raise UnitException("Must be UnitSet")

        lost = self.missing(units=concrete_set.units)
        gained = concrete_set.missing(units=self.units)

        for u in gained.values():
            u.set_state(UnitState.DEFAULT)

        if len(gained) == 0 and len(lost) == 0:
            self.logger.debug("Updating properties on Controller side for modify or extend")
            self.update(units=concrete_set.units)

        self.remove_from_dict(units=lost, configure=configure)
        self.add_from_dict(units=gained)
        if concrete_set:
            self.transfer_in_units(units=gained)

    def update(self, *, units: dict):
        """
        Update the units
        @param units units to be updated
        """
        for u in units.values():
            u.set_reservation(self.reservation)
            u.set_slice_id(self.reservation.get_slice_id())
            u.set_actor_id(self.plugin.get_actor().get_guid())
            self.plugin.update_props(reservation=self.reservation, u=u)

    def clone(self):
        result = UnitSet(plugin=self.plugin, units=self.units.copy())
        result.is_closed = self.is_closed
        result.is_fresh = self.is_fresh
        result.reservation = self.reservation
        return result

    def clone_empty(self) -> UnitSet:
        result = UnitSet(plugin=self.plugin)
        result.is_fresh = True
        return result

    def close(self):
        lost = self.units.copy()
        self.transfer_out_units(units=lost)
        self.is_closed = True

    def collect_released(self) -> IConcreteSet:
        result = None
        if self.released is not None and len(self.released) > 0:
            result = UnitSet(plugin=self.plugin, units=self.released)
            self.released = None

        return result

    def select_extract(self, *, count: int, victims: str) -> dict:
        """
        Extract specified victims
        @param count count
        @param victims token string identifying the victims
        """
        num_taken = 0
        taken = {}

        if victims is not None:
            for v in victims.split(" "):
                uid = ID(uid=v)
                if uid in self.units:
                    taken[uid] = self.units[uid]
                    num_taken += 1

        for u in self.units.values():
            if num_taken == count:
                break

            if u.get_id() not in taken:
                taken[u.get_id()] = u
                num_taken += 1

        return taken

    def get_notices(self) -> Notice:
        result = Notice()
        for u in self.units.values():
            n = u.get_notices()
            if not n.is_empty():
                result.add(msg=n.get_notice())
        return result

    def get_site_proxy(self) -> IAuthorityProxy:
        return None

    def get_units(self) -> int:
        return len(self.units)

    def holding(self, *, when: datetime) -> int:
        return self.get_units()

    def get_pending_count(self) -> int:
        """
        Get Pending Action Count
        """
        count = 0
        for u in self.units.values():
            if u.has_pending_action():
                count += 1

        return count

    def is_active(self) -> bool:
        return not self.is_fresh and self.reservation is not None and self.get_pending_count() == 0

    def modify(self, *, concrete_set: IConcreteSet, configure: bool):
        self.ensure_type(cset=concrete_set)

        for u in concrete_set.units.values():
            if u.get_id() in self.units:
                self.units[u.get_id()].set_modified(u)
                if configure:
                    self.modify_unit(u=self.units[u.get_id()])
            else:
                self.logger.warning("Modify for unit not present in seet: {}".format(u.get_id()))

    def probe(self):
        rel = None
        for u in self.units.values():
            if u.is_closed() or u.is_failed():
                if rel is None:
                    rel = {}
                rel[u.get_id()] = u

        if rel is not None:
            if self.released is None:
                self.released = rel
            else:
                for u in rel.values():
                    self.released[u.get_id()] = u
                    self.units.pop(u.get_id())

    def remove(self, *, concrete_set: IConcreteSet, configure: bool):
        self.ensure_type(cset=concrete_set)
        self.is_fresh = False

        self.remove_from_dict(units=configure.units, configure=configure)

    def remove_from_dict(self, *, units: dict, configure: bool):
        """
        Remove units from units dictionary
        @param units units to be removed
        @param configure flag to indicate if transfer_out to be triggerd or not
        """
        self.is_fresh = False
        for u in units.values():
            if u in self.units:
                self.units.pop(u)
            if configure:
                self.transfer_out(u=u)

    def setup(self, *, reservation: IReservation):
        self.reservation = reservation

    def validate_concrete(self, *, rtype: ResourceType, units: int, term: Term):
        if self.get_units() < units:
            raise UnitException("Insufficient units")

    def validate_incoming(self):
        """
        Validate incoming unit
        """

    def validate_outgoing(self):
        """
        Validate an outgoing unit
        """

    def modify_unit(self, *, u: Unit):
        """
        Modify a unit
        @param u unit
        """
        try:
            u.start_modify()
            self.plugin.modify(reservation=self.reservation, u=u)
        except Exception as e:
            self.fail(u=u, message="Modify for node failed", e=e)

    def restart_actions(self):
        """
        Restart actions
        """
        for u in self.units.values():
            if u.get_state() == UnitState.ACTIVE:
                self.logger.debug("Do nothing")
            elif u.get_state() == UnitState.CLOSING:
                u.decrement_sequence()
                self.transfer_out(u=u)
            elif u.get_state() == UnitState.PRIMING or u.get_state() == UnitState.DEFAULT:
                u.decrement_sequence()
                self.transfer_in(unit=u)
            elif u.get_state() == UnitState.MODIFYING:
                u.decrement_sequence()
                self.modify_unit(u=u)
            elif u.get_state() == UnitState.FAILED or u.get_state() == UnitState.CLOSED:
                self.logger.debug("Do nothing")

    def transfer_in(self, *, unit: Unit):
        """
        Transfer in a unit
        @param unit unit
        """
        try:
            if unit.start_prime():
                unit.set_reservation(reservation=self.reservation)
                unit.set_slice_id(slice_id=self.reservation.get_slice_id())
                unit.set_actor_id(actor_id=self.plugin.get_actor().get_guid())
                self.plugin.transfer_in(reservation=self.reservation, unit=unit)
            else:
                self.post(u=unit, message="Unit cannot be transfered., State={}".format(unit.get_state()))
        except Exception as e:
            self.fail(u=unit, message="Transfer in for node failed", e=e)

    def post(self, *, u: Unit, message: str):
        """
        Post the notice
        @param u unit
        @param message message
        """
        self.logger.error(message)
        u.add_notice(notice=message)

    def fail(self, *, u: Unit, message: str, e: Exception = None):
        """
        Fail a unit and log error message
        @param u unit
        @param message message
        @param e exception
        """
        self.logger.error(message)
        if e is not None:
            self.logger.error(e)

        u.fail(message=message, exception=e)

    def transfer_in_units(self, *, units: dict):
        """
        Transfer in the units
        @param units units to be transferred
        """
        for u in units.values():
            self.transfer_in(unit=u)

    def transfer_out(self, *, u: Unit):
        """
        Transfer a unit
        @param u unit to be transferred
        """
        if u.transfer_out_started:
            return

        try:
            u.start_close()
            self.plugin.transfer_out(reservation=self.reservation, unit=u)
        except Exception as e:
            self.fail(u=u, message="tranferOut error", e=e)

    def transfer_out_units(self, *, units: dict):
        """
        Transfer out units
        @param units units to be transferred
        """
        for u in units.values():
            self.transfer_out(u=u)

    def get_set(self) -> dict:
        """
        Get a copy of the units
        """
        return self.units.copy()

    def encode(self, *, protocol: str):
        try:
            encoded_unit = pickle.dumps(self)
            return encoded_unit
        except Exception as e:
            self.logger.error("Exception occurred while encoding {}".format(e))
        return None

    def decode(self, *, encoded, plugin: IBasePlugin):
        try:
            unit_set = pickle.loads(encoded)
            unit_set.plugin = plugin
            unit_set.logger = plugin.get_logger()
            return unit_set
        except Exception as e:
            self.logger.error("Exception occurred while decoding {}".format(e))
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum


class ErrorCodes(Enum):
    """
    Error code enumeration
    """
    ErrorInvalidArguments = 1
    ErrorInvalidActor = 2
    ErrorInvalidReservation = 3
    ErrorDatabaseError = 4
    ErrorInternalError = 5
    ErrorNoSuchSlice = 6
    ErrorNoSuchResourcePool = 7
    ErrorNoSuchReservation = 8
    ErrorNoSuchBroker = 9
    ErrorInvalidSlice = 10
    ErrorNoSuchActor = 11
    ErrorTransportFailure = 12
    ErrorTransportTimeout = 13
    ErrorNoSuchDelegation = 14


class Constants:
    """
    Constants
    """
    reservation_has_pending_operation = -200001
    extend_same_units = -1
    all_reservation_states = -1
    management_api_timeout_in_seconds = 120.0

    container_managment_object_id = "manager"

    property_pickle_properties = "properties"

    config_section_runtime = "runtime"
    property_conf_kafka_server = "kafka-server"
    property_conf_kafka_schema_registry = "kafka-schema-registry-url"
    property_conf_kafka_key_schema = "kafka-key-schema"
    property_conf_kafka_value_schema = "kafka-value-schema"
    property_conf_kafka_s_sl_ca_location = "kafka-ssl-ca-location"
    property_conf_kafka_ssl_certificate_location = "kafka-ssl-certificate-location"
    property_conf_kafka_ssl_key_location = "kafka-ssl-key-location"
    property_conf_kafka_ssl_key_password = "kafka-ssl-key-password"
    property_conf_kafka_security_protocol = "kafka-security-protocol"
    property_conf_kafka_group_id = "kafka-group-id"
    property_conf_kafka_sasl_producer_username = "kafka-sasl-producer-username"
    property_conf_kafka_sasl_producer_password = "kafka-sasl-producer-password"
    property_conf_kafka_sasl_consumer_username = "kafka-sasl-consumer-username"
    property_conf_kafka_sasl_consumer_password = "kafka-sasl-consumer-password"
    property_conf_kafka_sasl_mechanism = "kafka-sasl-mechanism"

    kafka_topic = "kafka-topic"
    name = "name"
    type = "type"
    guid = "guid"
    credmgr_host = "credmgr-host"

    property_class_name = "ObjectClassName"
    property_module_name = "ModuleName"
    property_id = "MOID"
    property_type_id = "MOTYPEID"
    property_actor_name = "MOActorName"
    property_proxies_length = "MOProxiesLength"
    property_proxies_prefix = "MOProxiesPrefix."
    property_proxies_protocol = ".protocol"
    property_proxies_class = ".class"
    property_proxies_module = ".module"

    property_conf_prometheus_rest_port = "prometheus.port"
    property_conf_controller_rest_port = "orchestrator.rest.port"
    property_conf_controller_create_wait_time_ms = "orchestrator.create.wait.time.ms"
    property_conf_controller_delay_resource_types = "orchestrator.delay.resource.types"

    property_substrate_file = "substrate.file"
    property_aggregate_resource_model = "AggregateResourceModel"

    property_conf_plugin_dir = "plugin-dir"

    config_logging_section = 'logging'

    config_section_pdp = 'pdp'

    property_conf_log_file = 'log-file'
    property_conf_log_level = 'log-level'
    property_conf_log_retain = 'log-retain'
    property_conf_log_size = 'log-size'
    property_conf_log_directory = 'log-directory'
    property_conf_logger = "logger"

    config_section_container = "container"
    property_conf_container_guid = "container.guid"

    config_section_time = "time"
    property_conf_time_start_time = "time.startTime"
    property_conf_time_cycle_millis = "time.cycleMillis"
    property_conf_time_manual = "time.manual"

    config_section_o_auth = "oauth"
    property_conf_o_auth_jwks_url = "oauth-jwks-url"
    property_conf_o_auth_token_public_key = "token-public-key"

    config_section_database = "database"
    property_conf_db_user = "db-user"
    property_conf_db_password = "db-password"
    property_conf_db_name = "db-name"
    property_conf_db_host = "db-host"

    config_section_neo4j = "neo4j"

    protocol_local = "local"
    protocol_kafka = "kafka"

    home_directory = '/usr/src/app/'
    superblock_location = home_directory + "state_recovery.lock"
    controller_lock_location = home_directory + "controller_recovery.lock"
    configuration_file = "/etc/fabric/actor/config/config.yaml"
    state_file_location = '/tmp/fabric_actor.tmp'

    test_directory = "/Users/komalthareja/renci/code/fabric/ControlFramework/fabric/actor/test/"
    test_broker_configuration_file = test_directory + "config/config.broker.yaml"
    test_net_am_configuration_file = test_directory + "config/config.net-am.yaml"
    test_vm_am_configuration_file = test_directory + "config/config.site.am.yaml"
    test_controller_configuration_file = test_directory + "config/config.orchestrator.yaml"

    elastic_time = "request.elasticTime"
    elastic_size = "request.elasticSize"

    pool_name = 'pool.name'
    pool_prefix = "pool."
    pools_count = "pools.count"
    query_action_discover_pools = "discover.pools"
    query_action = "query.action"
    query_response = "query.response"
    broker_query_model = "bqm"
    pool_type = "neo4j"

    config_handler = "config.handler"
    config_victims = "config.victims"
    config_image_guid = "config.image.guid"
    config_ssh_key_pattern = "config.ssh.user%d.keys"
    config_ssh_login_pattern = "config.ssh.user%d.login"
    config_ssh_sudo_pattern = "config.ssh.user%d.sudo"
    config_ssh_urn_pattern = "config.ssh.user%d.urn"
    config_ssh_num_logins = "config.ssh.numlogins"
    config_ssh_prefix = "config.ssh.user"
    config_ssh_key_suffix = ".keys"
    config_ssh_login_suffix = ".login"
    config_ssh_sudo_suffix = ".sudo"
    config_ssh_urn_suffix = ".urn"
    config_unit_tag = "config.unit.tag"

    resource_memory = "resource.memory"
    resource_cpu = "resource.cpu"
    resource_bandwidth = "resource.bandwidth"
    resource_num_cpu_cores = "resource.numCPUCores"
    resource_memory_capacity = "resource.memeoryCapacity"
    resource_storage_capacity = "resource.storageCapacity"
    resource_class_inventory_for_type = "resource.class.invfortype"
    resource_available_units = "resource.units.now"
    resource_start_iface = "resource.siface"
    resource_end_iface = "resource.eiface"
    resource_domain = "resource.domain"
    resource_neo4j_abstract_domain = "resource.neo4j.adomain"

    unit_management_ip = "unit.manage.ip"
    unit_management_port = "unit.manage.port"
    unit_manage_subnet = "unit.manage.subnet"
    unit_manage_gateway = "unit.manage.gateway"
    unit_data_subnet = "unit.data.subnet"
    unit_parent_host_name = "unit.parent.hostname"
    unit_host_name = "unit.hostname"
    unit_control = "unit.control"
    unit_memory = "unit.memory"
    unit_vlan_tag = "unit.vlan.tag"
    unit_vlan_qo_s_rate = "unit.vlan.qos.rate"
    unit_vlan_qo_s_burst_size = "unit.vlan.qos.burst.size"
    unit_eth_prefix = "unit.eth"
    unit_number_interface = "unit.number.interface"
    unit_modify_prop_message_suffix = ".message"
    unit_modify_prop_code_suffix = ".code"
    unit_modify_prop_prefix = "unit.modify."
    unit_lun_tag = "unit.target.lun"
    unit_storage_capacity = "unit.target.capacity"

    property_start_vlan = "vlan.tag.start"
    property_end_vlan = "vlan.tag.end"
    property_vlan_range_num = "vlan.range.num"

    property_start_lun = "lun.tag.start"
    property_end_lun = "lun.tag.end"
    property_lun_range_num = "lun.range.num"

    property_delegation_slice_id = 'dlg_slc_id'
    property_reservation_slice_id = 'rsv_slc_id'
    property_reservation_id = "rsv_resid"

    sasl_username = 'sasl.username'
    sasl_password = 'sasl.password'
    sasl_mechanism = 'sasl.mechanism'
    bootstrap_servers = 'bootstrap.servers'
    security_protocol = 'security.protocol'
    group_id = 'group.id'
    ssl_ca_location = 'ssl.ca.location'
    ssl_certificate_location = 'ssl.certificate.location'
    ssl_key_location = 'ssl.key.location'
    ssl_key_password = 'ssl.key.password'
    schema_registry_url = 'schema.registry.url'

    invalid_argument = "Invalid argument"
    uninitialized_state = "Uninitialized state"
    invalid_actor_state = "Invalid state, actor cannot receive calls"
    object_not_found = "{} not found in datatabase e: {}"
    failed_to_add = "Failed to add {} to database e: {}"
    failed_to_remove = "Failed to remove {} from database e: {}"
    not_specified_prefix = "Not {} specified"
    recovery = "[recovery]"
    issue_operation = "Issued {} request for reservation #{} State={}"
    invalid_pending_state = "Invalid pending state"
    restarting_actions = "Restarting configuration actions for reservation #{}"
    restarting_actions_complete = "Restarting configuration actions for reservation #{} complete"
    extend_ticket = "[extend ticket]"

    management_inter_actor_outbound_message = "Outbound [Kafka] [Management] Message {} written to {}"
    management_inter_actor_inbound_message = "Inbound [Kafka] [Management] Message received {}"
    management_api_timeout_occurred = "Management API Timeout Occurred"
    management_inter_actor_message_failed = "Outbound [Kafka] [Management] Message {} could not be written to {}"

    invalid_management_object_type = "Invalid Management Object type {}"
    no_pool = "there is no pool to satisfy this request"
    invalid_recovery_state = "This state should not be reached during recovery"

    invalid_state = "Invalid state"
    invalid_ip = "Invalid ip address: {}"
    exception_occurred = "Exception occurred {}"
    not_implemented = "Not Implemented"



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import threading
from datetime import datetime

from fabric.actor.core.apis.i_tick import ITick
from fabric.actor.core.kernel.tick import Tick


class KernelTick(Tick):
    """
    Timer object.
    """
    def __init__(self):
        super().__init__()
        self.timer = None
        self.current_cycle = -1
        self.to_tick = set()
        self.stopped_worker = threading.Event()

    def add_tickable(self, *, tickable: ITick):
        """
        Add object which needs to receive periodic tick
        """
        with self.lock:
            self.to_tick.add(tickable)

    def remove_tickable(self, *, tickable: ITick):
        """
        Remove tickable object
        """
        with self.lock:
            self.to_tick.remove(tickable)

    def start_worker(self):
        """
        Start timer thread
        """
        self.timer = threading.Thread(target=self.tick_notifier)
        self.timer.setDaemon(True)
        self.timer.setName("KernelTick")
        self.timer.start()

    def stop_worker(self):
        """
        Stop timer thread
        """
        self.stopped_worker.set()
        self.timer.join()

    def next_tick(self):
        """
        Generate a tick to the registered objects
        """
        with self.lock:
            now = datetime.utcnow()
            self.current_cycle = self.clock.cycle(when=now)
            self.logger.debug("Clock interrupt: now= {} cycle={}".format(now, self.current_cycle))
            if not self.manual and self.timer is None:
                return

            for t in self.to_tick:
                try:
                    self.logger.debug("Delivering external tick to {} cycle= {}".format(
                        t.get_name(), self.current_cycle))
                    t.external_tick(cycle=self.current_cycle)
                except Exception as e:
                    self.logger.error("Unexpected error while delivering tick notification for {} {}".format(
                        t.get_name(), e))

    def tick_notifier(self):
        """
        Timer main loop
        """
        while not self.stopped_worker.wait(timeout=self.cycle_millis / 1000):
            self.next_tick()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from datetime import datetime

from fabric.actor.core.apis.i_actor import ActorType
from fabric.actor.core.common.constants import ErrorCodes, Constants
from fabric.actor.core.common.exceptions import ManageException
from fabric.actor.core.manage.kafka.services.kafka_actor_service import KafkaActorService
from fabric.actor.core.manage.management_object import ManagementObject
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.util.resource_type import ResourceType
from fabric.message_bus.messages.extend_reservation_avro import ExtendReservationAvro
from fabric.message_bus.messages.get_actors_request_avro import GetActorsRequestAvro
from fabric.message_bus.messages.get_pool_info_request_avro import GetPoolInfoRequestAvro
from fabric.message_bus.messages.message import IMessageAvro
from fabric.message_bus.messages.reclaim_resources_avro import ReclaimResourcesAvro
from fabric.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric.message_bus.messages.result_pool_info_avro import ResultPoolInfoAvro
from fabric.message_bus.messages.result_proxy_avro import ResultProxyAvro
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.message_bus.messages.add_reservation_avro import AddReservationAvro
from fabric.message_bus.messages.add_reservations_avro import AddReservationsAvro
from fabric.message_bus.messages.claim_resources_avro import ClaimResourcesAvro
from fabric.actor.core.util.id import ID
from fabric.message_bus.messages.demand_reservation_avro import DemandReservationAvro
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric.message_bus.messages.result_string_avro import ResultStringAvro
from fabric.message_bus.messages.result_strings_avro import ResultStringsAvro


class KafkaClientActorService(KafkaActorService):
    def __init__(self):
        super().__init__()

    def process(self, *, message: IMessageAvro):
        callback_topic = message.get_callback_topic()
        result = None

        self.logger.debug("Processing message: {}".format(message.get_message_name()))

        if message.get_message_name() == IMessageAvro.ClaimResources:
            result = self.claim(request=message)
        elif message.get_message_name() == IMessageAvro.ReclaimResources:
            result = self.reclaim(request=message)
        elif message.get_message_name() == IMessageAvro.AddReservation:
            result = self.add_reservation(request=message)
        elif message.get_message_name() == IMessageAvro.AddReservations:
            result = self.add_reservations(request=message)
        elif message.get_message_name() == IMessageAvro.DemandReservation:
            result = self.demand_reservation(request=message)
        elif message.get_message_name() == IMessageAvro.GetActorsRequest:
            result = self.get_brokers(request=message)
        elif message.get_message_name() == IMessageAvro.GetPoolInfoRequest:
            result = self.get_pool_info(request=message)
        elif message.get_message_name() == IMessageAvro.ExtendReservation:
            result = self.extend_reservation(request=message)
        else:
            super().process(message=message)
            return

        if callback_topic is None:
            self.logger.debug("No callback specified, ignoring the message")

        if self.producer.produce_sync(callback_topic, result):
            self.logger.debug("Successfully send back response: {}".format(result.to_dict()))
        else:
            self.logger.debug("Failed to send back response: {}".format(result.to_dict()))

    def add_reservation(self, *, request: AddReservationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None or request.get_reservation() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.add_reservation(reservation=request.reservation_obj, caller=auth)
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def add_reservations(self, *, request: AddReservationsAvro) -> ResultStringsAvro:
        result = ResultStringsAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None or request.get_reservation() is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result.status = mo.add_reservations(reservations=request.reservation_list, caller=auth)
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def demand_reservation(self, *, request: DemandReservationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()

        try:
            if request.guid is None or (request.reservation_id is None and request.reservation_obj is None):
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)

                return result

            if request.reservation_obj is not None and request.reservation_id is not None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)

                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            if request.reservation_id is not None:
                result.status = mo.demand_reservation_rid(rid=ID(uid=request.reservation_id), caller=auth)
            else:
                result.status = mo.demand_reservation(reservation=request.reservation_obj, caller=auth)

            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_brokers(self, *, request: GetActorsRequestAvro) -> ResultProxyAvro:
        result = ResultProxyAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None or request.type != ActorType.Broker.name:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_brokers(caller=auth, id_token=request.get_id_token())
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_pool_info(self, *, request: GetPoolInfoRequestAvro) -> ResultPoolInfoAvro:
        result = ResultPoolInfoAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            result = mo.get_pool_info(broker=ID(uid=request.broker_id), caller=auth, id_token=request.get_id_token())
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def extend_reservation(self, *, request: ExtendReservationAvro) -> ResultStringAvro:
        result = ResultStringAvro()
        try:
            if request.guid is None or request.reservation_id is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)

                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            end_time = ActorClock.from_milliseconds(milli_seconds=request.end_time)
            rtype = None
            if request.new_resource_type is not None:
                rtype = ResourceType(resource_type=request.new_resource_type)

            result.status = mo.extend_reservation(rid=ID(uid=request.reservation_id), new_end_time=end_time,
                                                  new_units=request.new_units, new_resource_type=rtype,
                                                  request_properties=request.request_properties,
                                                  config_properties=request.config_properties, caller=auth)
            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def claim(self, *, request: ClaimResourcesAvro):
        result = ResultDelegationAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None or request.broker_id is None or request.delegation_id is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            if mo is None:
                self.logger.debug("Management object could not be found: guid: {} auth: {}".format(request.guid, auth))
                result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                result.status.set_message(ErrorCodes.ErrorNoSuchBroker.name)
                return result

            result = mo.claim_delegations(broker=ID(uid=request.broker_id), did=request.delegation_id, caller=auth,
                                          id_token=request.id_token)

            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            result.status.set_message(str(e))
            result.status.set_details(traceback.format_exc())

        return result

    def reclaim(self, *, request: ReclaimResourcesAvro):
        result = ResultDelegationAvro()
        result.status = ResultAvro()
        try:
            if request.guid is None or request.broker_id is None or request.delegation_id is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

            auth = Translate.translate_auth_from_avro(auth_avro=request.auth)
            mo = self.get_actor_mo(guid=ID(uid=request.guid))

            if mo is None:
                self.logger.debug("Management object could not be found: guid: {} auth: {}".format(request.guid, auth))
                result.status.set_code(ErrorCodes.ErrorNoSuchBroker.value)
                result.status.set_message(ErrorCodes.ErrorNoSuchBroker.name)
                return result

            result = mo.reclaim_delegations(broker=ID(uid=request.broker_id), did=request.delegation_id, caller=auth,
                                            id_token=request.id_token)

            result.message_id = request.message_id

        except Exception as e:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            result.status.set_message(str(e))
            result.status.set_details(traceback.format_exc())

        return result


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import pickle
import time
import unittest

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_actor_identity import IActorIdentity
from fabric.actor.core.apis.i_authority import IAuthority
from fabric.actor.core.apis.i_authority_reservation import IAuthorityReservation
from fabric.actor.core.apis.i_client_reservation import IClientReservation
from fabric.actor.core.apis.i_database import IDatabase
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import AuthorityException
from fabric.actor.core.core.ticket import Ticket
from fabric.actor.core.kernel.authority_reservation_factory import AuthorityReservationFactory
from fabric.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.plugins.config.config import Config
from fabric.actor.core.plugins.substrate.substrate import Substrate
from fabric.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.util.update_data import UpdateData
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.test.base_test_case import BaseTestCase
from fabric.actor.test.controller_callback_helper import ControllerCallbackHelper


class AuthorityPolicyTest(BaseTestCase):
    DonateStartCycle = 10
    DonateEndCycle = 100
    DonateUnits = 123
    Type = ResourceType(resource_type="1")
    Label = "test resource"

    TicketStartCycle = DonateStartCycle + 1
    TicketEndCycle = TicketStartCycle + 10
    TicketNewEndCycle = TicketEndCycle + 10
    TicketUnits = 1

    logger = logging.getLogger('AuthorityPolicyTest')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="actor.log")
    logger.setLevel(logging.INFO)

    authority = None

    def make_actor_database(self) -> IDatabase:
        db = SubstrateActorDatabase(user=self.db_user, password=self.db_pwd, database=self.db_name, db_host=self.db_host,
                                    logger=self.logger)
        return db

    def make_plugin(self):
        config = Config()
        plugin = Substrate(actor=None, db=None, config=config)
        return plugin

    def get_authority(self, name: str = BaseTestCase.authority_name, guid: ID = BaseTestCase.authority_guid):
        db = self.get_container_database()
        db.reset_db()
        authority = super().get_authority()
        authority.set_recovered(value=True)
        Term.set_clock(authority.get_actor_clock())
        return authority

    def get_source(self, units: int, rtype: ResourceType, term: Term, actor: IActor, slice_obj: ISlice):
        raise AuthorityException("Not implemented")

    def get_ticket(self, units: int, rtype: ResourceType, term: Term, source: IClientReservation, actor: IActor, holder: ID) -> Ticket:
        src_ticket = source.get_resources().get_resources().get_ticket()
        properties = src_ticket.get_delegation().get_properties()

        delegation = actor.get_plugin().get_ticket_factory().make_delegation(units=units, term=term, rtype=rtype, properties=properties, holder=holder)
        ticket = actor.get_plugin().get_ticket_factory().make_ticket(delegation=delegation, source=src_ticket)
        cs = Ticket(ticket=ticket, plugin=actor.get_plugin(), authority=None)
        return cs

    def get_request_slice(self):
        return SliceFactory.create(slice_id=ID(), name="test-slice", data=ResourceData())

    def get_request(self, units: int, rtype: ResourceType, term: Term, ticket: Ticket):
        rset = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())
        rset.set_resources(cset=ticket)
        slice_object = self.get_request_slice()
        return AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=slice_object, rid=ID())

    def get_request_from_request(self, request: IAuthorityReservation, units: int, rtype: ResourceType, term: Term, ticket: Ticket):
        rset = ResourceSet(units=units, rtype=rtype, rdata=ResourceData())
        rset.set_resources(cset=ticket)
        result = AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=request.get_slice(),
                                                    rid=request.get_reservation_id())
        result.set_sequence_in(sequence=request.get_sequence_in() + 1)
        return result

    def check_before_donate(self, authority: IAuthority):
        return

    def check_after_donate(self, authority: IAuthority, source: IClientReservation):
        return

    def get_donate_source(self, actor: IActor) -> IClientReservation:
        slice_object = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        slice_object.set_inventory(value=True)
        actor.register_slice(slice_object=slice_object)
        start = actor.get_actor_clock().cycle_start_date(cycle=self.DonateStartCycle)
        end = actor.get_actor_clock().cycle_end_date(cycle=self.DonateEndCycle)
        term = Term(start=start, end=end)

        source = self.get_source(self.DonateUnits, self.Type, term, actor, slice_object)
        actor.register(reservation=source)
        return source

    def check_before_donate_set(self, authority: IAuthority):
        return

    def check_after_donate_set(self, authority: IAuthority, rset: ResourceSet):
        return

    def get_donate_set(self, authority: IAuthority) -> ResourceSet:
        return ResourceSet(units=self.DonateUnits, rtype=self.Type, rdata=ResourceData())

    def get_redeem_request(self, authority: IAuthority, source: IClientReservation, identity: IActorIdentity):
        req_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketStartCycle)
        req_end = authority.get_actor_clock().cycle_end_date(cycle=self.TicketEndCycle)
        req_term = Term(start=req_start, end=req_end)
        ticket = self.get_ticket(units=self.TicketUnits, rtype=self.Type, term=req_term, source=source,
                                 actor=authority, holder=identity.get_guid())
        request = self.get_request(self.TicketUnits, self.Type, req_term, ticket)
        return request

    def get_extend_lease_request(self, authority: IAuthority, source: IClientReservation, identity: IActorIdentity, request: IAuthorityReservation):
        req_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketStartCycle)
        req_new_start = authority.get_actor_clock().cycle_start_date(cycle=self.TicketEndCycle + 1)
        req_end = authority.get_actor_clock().cycle_end_date(cycle=self.TicketNewEndCycle)
        req_term = Term(start=req_start, end=req_end, new_start=req_new_start)
        ticket = self.get_ticket(self.TicketUnits, self.Type, req_term, source, authority, identity.get_guid())
        new_request = self.get_request_from_request(request, self.TicketUnits, self.Type, req_term, ticket)
        return new_request

    def check_incoming_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation, udd: UpdateData):
        self.assertIsNotNone(incoming)

        rset = incoming.get_resources()

        self.assertIsNotNone(rset)

        self.assertEqual(request.get_requested_units(), rset.get_units())

        self.assertEqual(incoming.get_term(), request.get_requested_term())

    def check_incoming_close_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation, udd: UpdateData):
        self.assertIsNotNone(incoming)

        rset = incoming.get_resources()

        self.assertIsNotNone(rset)

        self.assertEqual(request.get_requested_units(), rset.get_units())

    def external_tick(self, site: IAuthority, cycle: int):
        site.external_tick(cycle=cycle)
        while site.get_current_cycle() != cycle:
            time.sleep(0.001)

    def check_incoming_extend_lease(self, authority: IAuthority, request: IAuthorityReservation, incoming: IReservation, update_data: UpdateData):
        self.assertIsNotNone(incoming)
        self.assertEqual(ReservationStates.Active, incoming.get_state())
        self.assertEqual(ReservationPendingStates.None_, incoming.get_pending_state())
        rset = incoming.get_resources()
        self.assertIsNotNone(rset)
        self.assertEqual(request.get_requested_units(), rset.get_units())
        self.assertEqual(incoming.get_term(), request.get_requested_term())

    def test_a_create(self):
        authority = self.get_authority()
        self.assertIsNotNone(authority)

    def test_b_donate(self):
        site = self.get_authority()
        policy = site.get_policy()
        self.check_before_donate(site)
        source = self.get_donate_source(site)
        policy.donate_reservation(reservation=source)
        self.check_after_donate(site, source)

    def test_c_donate_set(self):
        site = self.get_authority()
        policy = site.get_policy()
        self.check_before_donate_set(site)
        rset = self.get_donate_set(site)
        policy.donate(resources=rset)
        self.check_after_donate_set(site, rset)

    def test_d_redeem(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        authority_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())

        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=authority_proxy)

        source = self.get_donate_source(site)
        policy.donate_reservation(reservation=source)

        source_set = self.get_donate_set(site)
        policy.donate(resources=source_set)

        request = self.get_redeem_request(site, source, proxy)

        class UpdateLeaseHandler(ControllerCallbackHelper.IUpdateLeaseHandler):
            def __init__(self, parent):
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent

            def handle_update_lease(self, reservation: IReservation, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(site, request, reservation, update_data)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    self.parent.assertTrue(site.get_current_cycle() >= AuthorityPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(site, request, reservation, update_data)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.invalid_state)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self)
        proxy.set_update_lease_handler(handler=handler)

        self.external_tick(site, 0)
        print("Redeeming request...")
        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())

        for cycle in range(1, self.DonateEndCycle):
            self.external_tick(site, cycle)

        handler.check_termination()

    def test_e_extend_lease(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        auth_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=auth_proxy)

        source = self.get_donate_source(site)
        policy.donate_reservation(reservation=source)
        source_set = self.get_donate_set(site)
        policy.donate(resources=source_set)
        request = self.get_redeem_request(site, source, proxy)
        extend = self.get_extend_lease_request(site, source, proxy, request)

        class UpdateLeaseHandler(ControllerCallbackHelper.IUpdateLeaseHandler):
            def __init__(self, parent):
                self.waiting_for_extend_lease = False
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent

            def handle_update_lease(self, reservation: IReservation, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(site, request, reservation, update_data)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    self.parent.assertTrue(site.get_current_cycle() >= AuthorityPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(site, request, reservation, update_data)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.invalid_state)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self)

        proxy.set_update_lease_handler(handler=handler)
        self.external_tick(site, 0)
        print("Redeeming request")
        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())
        for cycle in range(self.DonateEndCycle):
            if cycle == self.TicketEndCycle - 50:
                print("Extending Lease")
                site.extend_lease(reservation=extend, caller=proxy.get_identity())
            self.external_tick(site, cycle)
        handler.check_termination()

    def test_f_close(self):
        site = self.get_authority()
        policy = site.get_policy()
        controller = self.get_controller()
        proxy = ControllerCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        authority_proxy = ControllerCallbackHelper(name=site.get_name(), guid=site.get_guid())

        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=authority_proxy)

        source = self.get_donate_source(site)
        policy.donate_reservation(reservation=source)

        source_set = self.get_donate_set(site)
        policy.donate(resources=source_set)

        request = self.get_redeem_request(site, source, proxy)

        class UpdateLeaseHandler(ControllerCallbackHelper.IUpdateLeaseHandler):
            def __init__(self, parent):
                self.waiting_for_lease = True
                self.waiting_for_close = False
                self.parent = parent

            def handle_update_lease(self, reservation: IReservation, update_data: UpdateData, caller: AuthToken):
                if self.waiting_for_lease:
                    self.parent.check_incoming_lease(authority=site, request=request, incoming=reservation,
                                                     udd=update_data)
                    self.waiting_for_lease = False
                    self.waiting_for_close = True
                elif self.waiting_for_close:
                    print(site.get_current_cycle())
                    self.parent.assertTrue(site.get_current_cycle() < AuthorityPolicyTest.TicketEndCycle)
                    self.parent.check_incoming_close_lease(authority=site, request=request, incoming=reservation,
                                                           udd=update_data)
                    self.waiting_for_close = False
                else:
                    raise AuthorityException(Constants.invalid_state)

            def check_termination(self):
                self.parent.assertFalse(self.waiting_for_lease)
                self.parent.assertFalse(self.waiting_for_close)

        handler = UpdateLeaseHandler(self)
        proxy.set_update_lease_handler(handler=handler)

        self.external_tick(site, 0)
        print("Redeeming request...")
        site.redeem(reservation=request, callback=proxy, caller=proxy.get_identity())

        for cycle in range(1, self.DonateEndCycle):
            if cycle == self.TicketEndCycle - 5:
                site.close(reservation=request)
            self.external_tick(site, cycle)

        handler.check_termination()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_delegation import IDelegation

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_broker_proxy import IBrokerProxy
    from fabric.actor.core.apis.i_reservation import IReservation
    from fabric.actor.core.apis.i_slice import ISlice
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.resource_type import ResourceType


class IDatabase:
    """
    IDatabase is the base database layer interface. It specifies methods for managing slices, reservations,
    broker proxies, and configuration mapping files.
    """

    @abstractmethod
    def actor_added(self):
        """
        Performs initialization actions as a result of the actor being
        added to the container.

        @throws Exception in case of error
        """

    @abstractmethod
    def add_broker(self, *, broker: IBrokerProxy):
        """
        Adds a new broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def add_reservation(self, *, reservation: IReservation):
        """
        Adds a new record to the database representing this reservation
        object.

        @param reservation reservation

        @throws Exception in case of error
        """

    @abstractmethod
    def add_slice(self, *, slice_object: ISlice):
        """
        Adds a new record to the database representing this slice
        object.

        @param slice_object Slice object

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_broker(self, *, broker: IBrokerProxy):
        """
        Removes the specified broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_reservation(self, *, rid: ID):
        """
        Removes the corresponding reservation object.

        @param rid reservation id

        @throws Exception in case of error
        """

    @abstractmethod
    def remove_slice(self, *, slice_id: ID):
        """
        Removes the corresponding database slice record.

        @param slice_id slice name

        @throws Exception in case of error
        """

    @abstractmethod
    def set_actor_name(self, *, name: str):
        """
        Sets the name of the actor this database belongs to.

        @param name actor name
        """

    @abstractmethod
    def update_broker(self, *, broker: IBrokerProxy):
        """
        Updates the specified broker proxy record.

        @param broker broker proxy

        @throws Exception in case of error
        """

    @abstractmethod
    def update_reservation(self, *, reservation: IReservation):
        """
        Updates the corresponding reservation object.

        @param reservation reservation

        @throws Exception in case of error
        """

    @abstractmethod
    def update_slice(self, *, slice_object: ISlice):
        """
        Updates the corresponding database slice record.

        @param slice_object slice object

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations(self) -> list:
        """
        Retrieves the reservations.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservation(self, *, rid: ID) -> dict:
        """
        Retrieves the specified reservation record.

        @param rid Reservation identifier

        @return dict of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations_by_slice_id(self, *, slice_id: ID) -> list:
        """
        Retrieves the specified reservations for a slice

        @param slice_id slice_id

        @return list of properties for reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations_by_slice_id_state(self, *, slice_id: ID, state: int) -> list:
        """
        Retrieves the specified reservations for a slice in a specific state

        @param slice_id slice_id
        @param state state

        @return list of properties for reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_client_reservations(self) -> list:
        """
        Retrieves the client reservations

        @return list of properties for reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_client_reservations_by_slice_id(self, *, slice_id: ID) -> list:
        """
        Retrieves the client reservations

        @param slice_id slice_id
        @return list of properties for reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_slice(self, *, slice_id: ID) -> dict:
        """
        Retrieves the specified slice record.

        @param slice_id slice id

        @return dict of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_slice_by_resource_type(self, *, rtype: ResourceType) -> dict:
        """
        Retrieves the specified slice record.

        @param rtype resource type

        @return dict of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_slices(self) -> list:
        """
        Retrieves all slice records.

        @return a list containing one or more properties dicts representing
                serialized slices

        @throws Exception in case of error
        """

    @abstractmethod
    def get_inventory_slices(self) -> list:
        """
        Retrieves all inventory slice records.

        @return a list containing one or more properties dicts representing
                serialized slices

        @throws Exception in case of error
        """

    @abstractmethod
    def get_client_slices(self) -> list:
        """
        Retrieves all client slice records.

        @return a list containing one or more properties dicts representing
                serialized slices

        @throws Exception in case of error
        """

    @abstractmethod
    def initialize(self):
        """
        Initializes the object.

        @throws Exception if initialization fails
        """

    @abstractmethod
    def get_holdings(self) -> list:
        """
        Retrieves all reservations representing resources held by this
        actor Broker/Controller.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_holdings_by_slice_id(self, *, slice_id: ID) -> list:
        """
        Retrieves all reservations representing resources held by this
        actor Broker/Controller.
        @param slice_id sliceId

        @return vector of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_broker_reservations(self) -> list:
        """
        Retrieves all reservations for which this actor acts as a
        broker.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_authority_reservations(self) -> list:
        """
        Retrieves all reservations for which this actor acts as a site.

        @return list of properties

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations_by_state(self, *, state: int) -> list:
        """
        Retrieves the reservations in a specific state

        @param state state

        @return list of properties for reservations

        @throws Exception in case of error
        """

    @abstractmethod
    def get_reservations_by_rids(self, *, rid: list):
        """
        Retrieves the specified reservation records.
        The order in the return vector is the same order as @rids
        @param rids rids
        @return list of properties
        @throws Exception in case of error
        """

    @abstractmethod
    def add_delegation(self, *, delegation: IDelegation):
        """
        Add delegation
        @params delegation: delegation
        """

    @abstractmethod
    def update_delegation(self, *, delegation: IDelegation):
        """
        Update delegation
        @params delegation: delegation
        """

    @abstractmethod
    def remove_delegation(self, *, dlg_graph_id: ID):
        """
        Remove delegation
        @params dlg_graph_id: dlg_graph_id
        """

    @abstractmethod
    def get_delegation(self, *, dlg_graph_id: ID) -> dict:
        """
        Get Delegation
        @params dlg_graph_id: dlg_graph_id
        """

    def get_delegations(self) -> list:
        """
        Get delegations
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ReservationException
from fabric.actor.core.util.id import ID


class ReservationFactory:
    @staticmethod
    def create_instance(*, properties: dict, actor: IActor, slice_obj: ISlice, logger) -> IReservation:
        """
        Creates and initializes a new reservation from a saved
        properties list.

        @param properties properties dict

        @return reservation instance

        @throws Exception in case of error
        """
        if Constants.property_pickle_properties not in properties:
            raise ReservationException(Constants.invalid_argument)

        serialized_reservation = properties[Constants.property_pickle_properties]
        deserialized_reservation = pickle.loads(serialized_reservation)
        deserialized_reservation.restore(actor=actor, slice_obj=slice_obj, logger=logger)
        return deserialized_reservation

    @staticmethod
    def get_reservation_id(*, properties: dict) -> ID:
        """
        Extracts the reservation identifier from the properties list.

        @param properties properties list

        @return reservation identifier

        @throws Exception if the properties list does not contain a reservation
                identifier
        """
        return ID(uid=properties[Constants.property_reservation_id])

    @staticmethod
    def get_slice_id(*, properties: dict) -> int:
        """
        Extracts the slice id from the properties list.

        @param properties properties list

        @return slice id

        @throws Exception if the properties list does not contain a slice name
        """
        slice_id = properties.get(Constants.property_reservation_slice_id, None)
        if slice_id is not None:
            return int(slice_id)
        return None

    @staticmethod
    def get_delegation_slice_id(*, properties: dict) -> int:
        """
        Extracts the slice id from the properties list.

        @param properties properties list

        @return slice id

        @throws Exception if the properties list does not contain a slice name
        """
        slice_id = properties.get(Constants.property_delegation_slice_id, None)
        if slice_id is not None:
            return int(slice_id)
        return None



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import unittest

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_controller_policy import IControllerPolicy
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.kernel.controller_reservation_factory import ControllerReservationFactory
from fabric.actor.core.kernel.reservation_states import ReservationStates
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.test.base_test_case import BaseTestCase
from fabric.actor.test.core.policy.controller_simple_policy_test_wrapper import ControllerSimplePolicyTestWrapper
from fabric.actor.test.core.policy.controller_test_wrapper import ControllerTestWrapper


class ControllerSimplePolicyTest(BaseTestCase, unittest.TestCase):
    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_controller_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def get_controller(self, name: str = BaseTestCase.controller_name, guid: ID = BaseTestCase.controller_guid):
        db = self.get_container_database()
        db.reset_db()
        controller = super().get_controller()
        controller.set_recovered(value=True)
        Term.set_clock(controller.get_actor_clock())
        return controller

    def get_controller_instance(self) -> IActor:
        return ControllerTestWrapper()

    def get_controller_policy(self) -> IControllerPolicy:
        policy = ControllerSimplePolicyTestWrapper()
        return policy

    def test_a_create(self):
        controller = self.get_controller()
        for i in range(1, 101):
            controller.external_tick(cycle=i)

    def test_b_demand(self):
        controller = self.get_controller()
        clock = controller.get_actor_clock()
        Term.clock = clock
        resources = ResourceSet(units=1, rtype=ResourceType(resource_type="1"), rdata=ResourceData())
        slice_obj = SliceFactory.create(slice_id=ID(), name="myslice")
        controller.register_slice(slice_object=slice_obj)

        start = 5
        end = 10

        term = Term(start=clock.cycle_start_date(cycle=start), end=clock.cycle_end_date(cycle=end))

        reservation = ControllerReservationFactory.create(rid=ID(), resources=resources, term=term, slice_object=slice_obj)
        reservation.set_renewable(renewable=False)
        controller.register(reservation=reservation)
        controller.demand(rid=reservation.get_reservation_id())

        for i in range(1, end+3):
            controller.external_tick(cycle=i)
            while controller.get_current_cycle() != i:
                time.sleep(0.001)

            if (i >= start) and (i < (end-1)):
                self.assertTrue(reservation.get_state() == ReservationStates.Active)

            if i > end:
                self.assertTrue(reservation.get_state() == ReservationStates.Closed)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle

from fabric.actor.core.util.id import ID
from .delegation import Delegation
from ..apis.i_actor import IActor
from ..apis.i_delegation import IDelegation
from ..apis.i_slice import ISlice
from ..common.constants import Constants
from ..common.exceptions import DelegationException


class DelegationFactory:
    """
    Factory class to create delegation instances
    """
    @staticmethod
    def create(did: ID, slice_id: ID) -> IDelegation:
        """
        Create a delegation
        @param did delegation id
        @param slice_id slice id
        @return delegation
        """
        delegation = Delegation(dlg_graph_id=did, slice_id=slice_id)
        return delegation

    @staticmethod
    def create_instance(*, properties: dict, actor: IActor, slice_obj: ISlice, logger) -> IDelegation:
        """
        Creates and initializes a new delegation from a saved
        properties list.

        @param properties properties dict
        @param actor actor
        @param slice_obj slice_obj
        @param logger logger

        @return delegation instance

        @raises Exception in case of error
        """
        if Constants.property_pickle_properties not in properties:
            raise DelegationException(Constants.invalid_argument)

        serialized_delegation = properties[Constants.property_pickle_properties]
        deserialized_delegation = pickle.loads(serialized_delegation)
        deserialized_delegation.restore(actor=actor, slice_obj=slice_obj, logger=logger)
        return deserialized_delegation



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from fabric.orchestrator.swagger_server.models.success import Success  # noqa: E501
from fabric.orchestrator.swagger_server.test import BaseTestCase


class TestResourcesController(BaseTestCase):
    """ResourcesController integration test stubs"""

    def test_resources_get(self):
        """Test case for resources_get

        Retrieve a listing and description of available resources
        """
        response = self.client.open(
            '//resources',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import PolicyException
from fabric.actor.core.common.resource_pool_attribute_descriptor import ResourcePoolAttributeDescriptor, \
    ResourcePoolAttributeType
from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor
from fabric.actor.core.policy.simpler_units_inventory import SimplerUnitsInventory
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.reflection_utils import ReflectionUtils
from fabric.actor.core.util.resource_type import ResourceType

if TYPE_CHECKING:
    from fabric.actor.core.policy.inventory_for_type import InventoryForType
    from fabric.actor.core.apis.i_client_reservation import IClientReservation


class Inventory:
    def __init__(self):
        self.map = {}

    def contains_type(self, *, resource_type: ResourceType):
        if resource_type is None:
            raise PolicyException(Constants.invalid_argument)

        if resource_type in self.map:
            return True

        return False

    def get(self, *, resource_type: ResourceType) -> InventoryForType:
        if resource_type is None:
            raise PolicyException(Constants.invalid_argument)

        return self.map.get(resource_type, None)

    def remove(self, *, source: IClientReservation):
        """
        Removes the inventory derived from the specified source.
        @param source source reservation
        @return true if the inventory was update, false otherwise
        """
        rtype = source.get_type()

        if rtype in self.map:
            inv = self.map[rtype]
            if inv.source == source:
                self.map.pop(rtype)
                return True

        return False

    def get_new(self, *, reservation: IClientReservation):
        if reservation is None:
            raise PolicyException(Constants.invalid_argument)

        rtype = reservation.get_type()

        if rtype in self.map:
            raise PolicyException("There is already inventory for type: {}".format(rtype))

        properties = {}

        rset = reservation.get_resources()
        cset = rset.get_resources()
        ticket = cset.get_ticket()

        properties = ticket.get_properties()
        properties = PropList.merge_properties(incoming=rset.get_resource_properties(), outgoing=properties)
        rpd = ResourcePoolDescriptor()
        rpd.reset(properties=properties)

        desc_attr = rpd.get_attribute(key=Constants.resource_class_inventory_for_type)
        inv = None
        if desc_attr is not None:
            module_name, class_name = desc_attr.get_value().rsplit(".", 1)
            inv = ReflectionUtils.create_instance(module_name=module_name, class_name=class_name)
        else:
            inv = SimplerUnitsInventory()

        inv.set_type(rtype=rtype)
        inv.set_descriptor(rpd=rpd)
        inv.donate(source=reservation)

        self.map[rtype] = inv
        return inv

    def get_inventory(self) -> dict:
        return self.map

    def get_resource_pools(self) -> dict:
        result = {}
        count = 0
        for inv in self.map.values():
            rpd = inv.get_descriptor().clone()
            attr = ResourcePoolAttributeDescriptor()
            attr.set_type(rtype=ResourcePoolAttributeType.INTEGER)
            attr.set_key(value=Constants.resource_available_units)
            attr.set_value(value=str(inv.get_free()))
            rpd.add_attribute(attribute=attr)
            result = rpd.save(properties=result, prefix=Constants.pool_prefix + str(count) + ".")
            count += 1

        result[Constants.pools_count] = str(len(self.map))
        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import FrameworkException


class UpdateData:
    """
    UpdateData wraps state passed with a ticket or lease update. Server-side code registers status
    of operations as they execute; client-side code queries for status, and may also post status.
    """
    def __init__(self, *, message: str = None):
        self.message = message
        self.events = None
        self.failed = False
        if self.message is None:
            self.message = ""

    def absorb(self, *, other):
        """
        Merges passed UpdateData into this. Posted events are extracted
        and merged into this. The status message from this is overwritten with
        the status message from the absorbed UpdateData.
        """
        if not isinstance(other, UpdateData):
            raise FrameworkException(Constants.invalid_argument)

        self.post(event=other.events)
        if self.message is not None:
            self.post(event=self.message)
        self.message = other.message
        self.failed = other.failed

    def clear(self):
        """
        Clears all events.
        """
        self.events = None

    def error(self, *, message: str):
        """
        Indicates that an error has occurred.

        @param message error message
        """
        self.message = message
        self.failed = True

    def is_failed(self) -> bool:
        """
        Checks if the operation represented by the object has failed.

        @return true if the operation has failed
        """
        return self.failed

    def get_events(self) -> str:
        """
        Returns all events stored in the object.

        @return list of events. Event items are separated by "\n".
        """
        return self.events

    def get_message(self) -> str:
        """
        Returns the message attached to the object.

        @return message
        """
        return self.message

    def post(self, *, event: str):
        """
        Posts a human-readable string describing an event that the user
        may wish to know about. If the object already contains messages, the
        new message is prepended to the existing messages. Messages are
        separated using "\n".

        @param event message describing event
        """
        if self.events is None:
            self.events = event
        else:
            self.events = "{}\n{}".format(event, self.events)

    def post_error(self, *, event: str):
        """
        Posts a human-readable string describing an event that the user
        may wish to know about, and also marks the UpdateData in a failed
        state.

        @param event message describing event
        """
        self.post(event=event)
        self.error(message=event)

    def successful(self):
        """
        Checks if the operation represented by the object has succeeded.

        @return true if the operation has succeeded 
        """
        return not self.failed



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy


class CallbackRegistry:
    def __init__(self):
        self.protocols = {}

    def clear(self):
        for protocol in self.protocols.values():
            protocol.clear()
        self.protocols.clear()

    def get_callback(self, *, protocol: str, actor_name: str):
        protocol_table = self.protocols.get(protocol, None)

        if protocol_table is None:
            return None

        return protocol_table.get(actor_name, None)

    def register_callback(self, *, callback: ICallbackProxy):
        protocol = callback.get_type()

        entry = self.protocols.get(protocol, None)

        if entry is None:
            entry = {}

        entry[callback.get_identity().get_name()] = callback
        self.protocols[protocol] = entry

    def unregister(self, *, actor_name: str):
        for protocol in self.protocols:
            if actor_name in protocol:
                protocol.pop(actor_name)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.util.id import ID
from fabric.actor.security.auth_token import AuthToken


class Client:
    def __init__(self):
        self.name = None
        self.guid = None
        self.type = None
        self.kafka_topic = None

    def __getstate__(self):
        state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def set_name(self, *, name: str):
        self.name = name

    def set_guid(self, *, guid: ID):
        self.guid = guid

    def set_type(self, *, client_type: str):
        self.type = client_type

    def get_name(self) -> str:
        return self.name

    def get_guid(self) -> ID:
        return self.guid

    def get_type(self) -> str:
        return self.type

    def get_kafka_topic(self) -> str:
        return self.kafka_topic

    def set_kafka_topic(self, *, kafka_topic: str):
        self.kafka_topic = kafka_topic

    def get_auth_token(self) -> AuthToken:
        return AuthToken(name=self.name, guid=self.guid)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
import traceback

from fabric.actor.core.apis.i_actor import IActor
from fabric.actor.core.apis.i_base_plugin import IBasePlugin
from fabric.actor.core.apis.i_callback_proxy import ICallbackProxy
from fabric.actor.core.apis.i_concrete_set import IConcreteSet
from fabric.actor.core.apis.i_proxy import IProxy
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.security.auth_token import AuthToken


class Proxy(IProxy):
    """
    Proxy class represents a stub to an actor. Proxies define a general interface, which is implementation
    independent and enables easy implementation of new communication protocols.
    """
    PropertyProxyActorName = "prx_name"

    @staticmethod
    def get_callback(*, actor: IActor, protocol: str) -> ICallbackProxy:
        """
        Obtains a callback for the specified actor
        @param actor actor
        @param protocol protocol
        @return ICallbackProxy
        """
        if actor is None:
            raise ProxyException(Constants.not_specified_prefix.format("actor"))

        callback = ActorRegistrySingleton.get().get_callback(protocol=protocol, actor_name=actor.get_name())
        if callback is None:
            raise ProxyException("Could not obtain callback proxy: protocol={}".format(protocol))
        return callback

    @staticmethod
    def get_proxy(*, properties: dict) -> IProxy:
        """
        Obtains a proxy object from the specified properties list. If a suitable
        proxy object has already been created and registered with the
        ActorRegistry, the already existing object is returned and
        no new object is created. Otherwise, the method creates the proxy object
        and registers it with the ActorRegistry
        @param properties Properties list representing the proxy
        @return IProxy
        @throws Exception in case of error
        """
        proxy_reload_from_db = pickle.loads(properties.get(Constants.property_pickle_properties))
        proxy_type = proxy_reload_from_db.get_type()
        name = proxy_reload_from_db.get_name()

        is_callback = proxy_reload_from_db.callback

        proxy = None
        if is_callback:
            proxy = ActorRegistrySingleton.get().get_callback(protocol=proxy_type, actor_name=name)
        else:
            proxy = ActorRegistrySingleton.get().get_proxy(protocol=proxy_type, actor_name=name)

        if proxy is None:
            proxy = Proxy.recover_proxy(proxy_reload_from_db=proxy_reload_from_db, register=True)
        else:
            proxy = Proxy.recover_proxy(proxy_reload_from_db=proxy_reload_from_db, register=False)
        return proxy

    @staticmethod
    def recover_proxy(*, proxy_reload_from_db: IProxy, register: bool) -> IProxy:
        """
        Creates a proxy list from a properties list representing the
        serialization of the proxy. Optionally, the resulting object may be
        registered with the ActorRegistry so that it becomes visible
        to the rest of the system.
        @param proxy_reload_from_db proxy_reload_from_db
        @param register If true, the resulting proxy is registered with the
                   container's ActorRegistry
        @return Proxy
        @throws Exception in case of error
        """

        from fabric.actor.core.container.globals import GlobalsSingleton
        proxy_reload_from_db.set_logger(GlobalsSingleton.get().get_logger())

        if register:
            if proxy_reload_from_db.callback:
                ActorRegistrySingleton.get().register_callback(callback=proxy_reload_from_db)
            else:
                ActorRegistrySingleton.get().register_proxy(proxy=proxy_reload_from_db)
        return proxy_reload_from_db

    @staticmethod
    def decode(*, encoded, plugin: IBasePlugin) -> IConcreteSet:
        try:
            decoded_resource = pickle.loads(encoded)
            print("Decoded object is of type={}".format(type(decoded_resource)))
            decoded_resource.restore(plugin=plugin, reservation=None)
            return decoded_resource
        except Exception as e:
            traceback.print_exc()
            print("Exception occurred while decoding {}".format(e))
        return None

    def __init__(self, *, auth: AuthToken = None):
        self.logger = None
        self.proxy_type = None
        self.callback = False
        self.actor_name = None
        if auth is not None:
            self.actor_name = auth.get_name()
        self.actor_guid = None
        if auth is not None:
            self.actor_guid = auth.get_guid()
        self.auth = auth

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = None

    def get_guid(self) -> ID:
        return self.actor_guid

    def get_identity(self) -> AuthToken:
        return self.auth

    def get_name(self) -> str:
        return self.actor_name

    def get_type(self) -> str:
        return self.proxy_type

    def set_logger(self, *, logger):
        self.logger = logger

    def get_logger(self):
        return self.logger

    def abstract_clone_authority(self, *, rset: ResourceSet) -> ResourceSet:
        """
        Clones the resource set, but without any of the concrete sets. Preserves
        only the configuration properties. This method should be used when
        sending a redeem/extend/close request to an authority.
        @param rset resource set
        @return a resources set that is a copy of the current but without any
                concrete sets.
        """
        new_resource_data = ResourceData()
        properties = rset.get_config_properties()

        if properties is None:
            properties = {}
        else:
            properties = properties.copy()

        temp = ResourceData.merge_properties(from_props=properties,
                                             to_props=new_resource_data.get_configuration_properties())
        new_resource_data.configuration_properties = temp
        return ResourceSet(units=rset.get_units(), rtype=rset.get_type(), rdata=new_resource_data)

    def abstract_clone_broker(self, *, rset: ResourceSet) -> ResourceSet:
        """
        Clones the resource set, but without any of the concrete sets. Preserves
        only the configuration properties. This method should be used when
        sending a redeem/extend/close request to an authority.
        @param rset resource set
        @return a resources set that is a copy of the current but without any
                concrete sets.
        """
        new_resource_data = ResourceData()
        properties = rset.get_request_properties()

        if properties is None:
            properties = {}
        else:
            properties = properties.copy()

        temp = ResourceData.merge_properties(from_props=properties, to_props=new_resource_data.get_request_properties())
        new_resource_data.request_properties = temp
        return ResourceSet(units=rset.get_units(), rtype=rset.get_type(), rdata=new_resource_data)

    @staticmethod
    def abstract_clone_return(rset: ResourceSet) -> ResourceSet:
        """
        Clones the resource set, but without any of the concrete sets. Preserves
        only the configuration properties. This method should be used when
        sending a redeem/extend/close request to an authority.
        @param rset resource set
        @return a resources set that is a copy of the current but without any
                concrete sets.
        """
        new_resource_data = ResourceData()
        properties = rset.get_resource_properties()

        if properties is None:
            properties = {}
        else:
            properties = properties.copy()

        temp = ResourceData.merge_properties(from_props=properties, to_props=new_resource_data.get_resource_properties())
        new_resource_data.resource_properties = temp
        return ResourceSet(units=rset.get_units(), rtype=rset.get_type(), rdata=new_resource_data)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import uuid


class ID:
    def __init__(self, *, uid: str = None):
        if uid is None:
            self.uid = uuid.uuid4().__str__()
        else:
            self.uid = uid

    def __str__(self):
        return str(self.uid)

    def print(self):
        print(str(self.uid))

    def __hash__(self):
        return hash(self.uid)

    def __lt__(self, other):
        if not isinstance(other, ID):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.uid < other.uid

    def __eq__(self, other):
        if not isinstance(other, ID):
            # don't attempt to compare against unrelated types
            return NotImplemented

        return self.uid == other.uid


def main():
    uid = ID()
    uid.print()
    print(uid.__hash__())
    uid1 = ID(uid="manager")
    uid1.print()
    print(uid1.__hash__())


if __name__ == "__main__":
    main()
